"""proxysql_tools module"""
import logging

__author__ = """TwinDB Development Team"""
__email__ = 'dev@twindb.com'
__version__ = '0.3.4'


LOG = logging.getLogger(__name__)


def setup_logging(logger, debug=False):     # pragma: no cover
    """Configure logging"""

    fmt_str = "%(asctime)s: %(levelname)s:" \
              " %(module)s.%(funcName)s():%(lineno)d: %(message)s"

    console_handler = logging.StreamHandler()

    console_handler.setFormatter(logging.Formatter(fmt_str))
    logger.handlers = []
    logger.addHandler(console_handler)
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)


def execute(conn, query, *args):
    """Execute query in connection"""
    cursor = conn.cursor()
    cursor.execute(query, *args)
    return cursor.fetchall()



"""Errors related to Galera"""


class GaleraClusterError(Exception):
    """Base Galera Cluster Error"""


class GaleraClusterNodeNotFound(GaleraClusterError):
    """Requested node not found"""


class GaleraClusterSyncedNodeNotFound(GaleraClusterNodeNotFound):
    """Cluster doesn't have a SYNCED node"""



import pytest

from proxysql_tools.galera.galera_cluster import GaleraCluster
from proxysql_tools.galera.galera_node import GaleraNode


def test_nodes():
    gc = GaleraCluster('foo:1')
    assert gc.nodes == [GaleraNode('foo', port=1)]


@pytest.mark.parametrize('cluster_host, result', [
    (
        "192.168.90.2:3306,192.168.90.3:3306,192.168.90.4:3306",
        [
            ('192.168.90.2', 3306),
            ('192.168.90.3', 3306),
            ('192.168.90.4', 3306),
        ]
    )
])
def test_split_cluster_host(cluster_host, result):
    assert GaleraCluster(cluster_host)._split_cluster_host(cluster_host) == result



import mock
from pymysql.cursors import DictCursor

from proxysql_tools.galera.galera_node import GaleraNode


def test_galera_node_init(galera_node):
    assert galera_node.host == 'foo'
    assert galera_node.port == 1234
    assert galera_node.user == 'bar'
    assert galera_node.password == 'xyz'


@mock.patch.object(GaleraNode, '_status')
def test_wsrep_cluster_state_uuid(mock_status, galera_node):
    """
    :param galera_node: GaleraNode instance
    :type galera_node: GaleraNode
    """
    assert galera_node.wsrep_cluster_state_uuid
    mock_status.assert_called_once_with('wsrep_cluster_state_uuid')


@mock.patch.object(GaleraNode, '_status')
def test_wsrep_cluster_status(mock_status, galera_node):
    """
    :param galera_node: GaleraNode instance
    :type galera_node: GaleraNode
    """
    assert galera_node.wsrep_cluster_status
    mock_status.assert_called_once_with('wsrep_cluster_status')


@mock.patch.object(GaleraNode, '_status')
def test_wsrep_local_state(mock_status, galera_node):
    """
    :param galera_node: GaleraNode instance
    :type galera_node: GaleraNode
    """
    mock_status.return_value = '4'
    assert galera_node.wsrep_local_state == 4
    mock_status.assert_called_once_with('wsrep_local_state')


@mock.patch.object(GaleraNode, 'execute')
def test_wsrep_cluster_name(mock_execute, galera_node):
    """
    :param galera_node: GaleraNode instance
    :type galera_node: GaleraNode
    """
    assert galera_node.wsrep_cluster_name
    mock_execute.assert_called_once_with('SELECT @@wsrep_cluster_name')


@mock.patch.object(GaleraNode, '_connect')
def test_execute(mock_connect, galera_node):
    """
    :param galera_node: GaleraNode instance
    :type galera_node: GaleraNode
    """
    mock_connection = mock.Mock()
    mock_connect.return_value.__enter__.return_value = mock_connection
    mock_cursor = mock.Mock()
    mock_connection.cursor.return_value = mock_cursor

    galera_node.execute('foo')
    mock_connect.assert_called_once_with()
    mock_connection.cursor.assert_called_once_with()
    mock_cursor.execute.assert_called_once_with('foo')
    mock_cursor.fetchall.assert_called_once_with()


@mock.patch('proxysql_tools.galera.galera_node.pymysql')
def test_connect(mock_pymysql, galera_node):
    with galera_node._connect() as conn:
        mock_pymysql.connect.assert_called_once_with(host='foo',
                                                     port=1234,
                                                     user='bar',
                                                     passwd='xyz',
                                                     cursorclass=DictCursor)


@mock.patch.object(GaleraNode, 'execute')
def test_status(mock_execute, galera_node):
    galera_node._status('foo')
    mock_execute.assert_called_once_with('SHOW GLOBAL STATUS LIKE %s', 'foo')



def test_galera_node_cluster_state_uuid(galera_node1):
    assert len(galera_node1.wsrep_cluster_state_uuid) == 36


def test_galera_node_cluster_status(galera_node1):
    assert galera_node1.wsrep_cluster_status == 'Primary'


def test_galera_node_wsrep_cluster_name(galera_node1):
    assert galera_node1.wsrep_cluster_name == 'LefredPXC'



import docker
import json
import socket
import time
import io
from ConfigParser import ConfigParser

from proxysql_tools import LOG
from proxysql_tools.galera.galera_node import GaleraNode


def docker_client():
    return docker.DockerClient(version="auto")


def docker_pull_image(image):
    """Pull the specified image using docker-py. This function will parse the
        result from docker-py and raise an exception if there is an error.

    :param str image: Name of the image to pull
    """
    client = docker_client()
    api = client.api

    response = api.pull(image)
    LOG.debug('Response: %s', response)


def get_unused_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    s.bind(('localhost', 0))
    _, port = s.getsockname()
    s.close()

    return port


def is_port_reachable(host, port):
    s = socket.socket()
    try:
        s.connect((host, port))
        return True
    except socket.error:
        return False


def eventually(func, *args, **kwargs):
    retries = kwargs.pop('retries', 90)
    sleep_time = kwargs.pop('sleep_time', 0.5)

    for i in xrange(retries):
        try:
            if func(*args, **kwargs):
                return
            else:
                LOG.info('Waiting for %s to return True', func)
                time.sleep(sleep_time)
        except Exception:
            time.sleep(sleep_time)
            continue

    raise EnvironmentError('Function %s never returned True' % func)


def shutdown_container(id):
    client = docker_client()
    api = client.api
    api.stop(id)

def create_percona_xtradb_cluster(container_image, container_labels,
                                  container_info, container_ports,
                                  network_name, cluster_name):
    client = docker_client()
    api = client.api

    # Pull the container image locally first
    docker_pull_image(container_image)

    bootstrapped = False
    cluster_join = ''
    for node in container_info:
        host_config = api.create_host_config(port_bindings={
            node['mysql_port']: get_unused_port()
        })

        networking_config = api.create_networking_config({
            network_name: api.create_endpoint_config(
                ipv4_address=node['ip']
            )
        })

        environment_vars = {
            'MYSQL_ROOT_PASSWORD': node['root_password'],
            'CLUSTER_JOIN': cluster_join,
            'CLUSTER_NAME': cluster_name,
            'XTRABACKUP_PASSWORD': 'xtrabackup'
        }

        container = api.create_container(
            image=container_image, name=node['name'],
            labels=container_labels, ports=container_ports,
            host_config=host_config, networking_config=networking_config,
            environment=environment_vars)
        api.start(container['Id'])

        node['id'] = container['Id']

        if not bootstrapped:
            cluster_join = node['ip']
            bootstrapped = True

        # We add a bit of delay to allow the node to startup before adding a
        # new node to the cluster
        time.sleep(5)

    return container_info


def wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_info):
    def check_started():
        for container_info in percona_xtradb_cluster_info:
            node = GaleraNode(host=container_info['ip'],
                              user='root',
                              password=container_info['root_password'])
            node.execute('SELECT 1')
        return True

    # Allow all the cluster nodes to startup.
    eventually(check_started, retries=20, sleep_time=4)


def proxysql_tools_config(proxysql_instance, cluster_host, cluster_port,
                          cluster_user, cluster_pass, hostgroup_writer,
                          hostgroup_reader, monitor_user, monitor_pass):
    config_contents = """
[proxysql]
host={proxy_host}
admin_port={proxy_port}
admin_username={proxy_user}
admin_password={proxy_pass}

monitor_username={monitor_user}
monitor_password={monitor_pass}

[galera]
cluster_host={cluster_host}:{cluster_port}
cluster_username={cluster_user}
cluster_password={cluster_pass}

load_balancing_mode=singlewriter

writer_hostgroup_id={writer_hostgroup}
reader_hostgroup_id={reader_hostgroup}
""".format(proxy_host=proxysql_instance.host, proxy_port=proxysql_instance.port,
           proxy_user=proxysql_instance.user,
           proxy_pass=proxysql_instance.password, monitor_user=monitor_user,
           monitor_pass=monitor_pass, cluster_host=cluster_host,
           cluster_port=cluster_port, cluster_user=cluster_user,
           cluster_pass=cluster_pass, writer_hostgroup=hostgroup_writer,
           reader_hostgroup=hostgroup_reader)

    config = ConfigParser()
    config.readfp(io.BytesIO(config_contents))
    return config

def proxysql_tools_config_2(proxysql_instance, cluster_nodes,
                             cluster_user, cluster_pass,
                             hostgroup_writer, hostgroup_reader, writer_blacklist,
                             monitor_user, monitor_pass):

    hosts = ','.join(cluster_nodes)
    config_contents = """
[proxysql]
host={proxy_host}
admin_port={proxy_port}
admin_username={proxy_user}
admin_password={proxy_pass}

monitor_username={monitor_user}
monitor_password={monitor_pass}

[galera]
cluster_host={hosts}
cluster_username={cluster_user}
cluster_password={cluster_pass}

load_balancing_mode=singlewriter
writer_blacklist={writer_blacklist}
writer_hostgroup_id={writer_hostgroup}
reader_hostgroup_id={reader_hostgroup}
""".format(proxy_host=proxysql_instance.host, proxy_port=proxysql_instance.port,
           proxy_user=proxysql_instance.user,
           proxy_pass=proxysql_instance.password, monitor_user=monitor_user,
           monitor_pass=monitor_pass, hosts=hosts,
           cluster_user=cluster_user,
           writer_blacklist=writer_blacklist,
           cluster_pass=cluster_pass, writer_hostgroup=hostgroup_writer,
           reader_hostgroup=hostgroup_reader)

    config = ConfigParser()
    config.readfp(io.BytesIO(config_contents))
    return config



"""Module describes GaleraNode class"""
from contextlib import contextmanager

import pymysql
from pymysql.cursors import DictCursor

from proxysql_tools import execute


class GaleraNodeState(object):  # pylint: disable=too-few-public-methods
    """State of Galera node http://bit.ly/2r1tUGB """
    PRIMARY = 1
    JOINER = 2
    JOINED = 3
    SYNCED = 4
    DONOR = 5


class GaleraNode(object):
    """
    GaleraNode class describes a single node in Galera Cluster.

    :param host: hostname of the node.
    :param port: port to connect to.
    :param user: MySQL username to connect to the node.
    :param password: MySQL password.
    """

    def __init__(self, host, port=3306, user='root', password=None):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    @property
    def wsrep_cluster_state_uuid(self):
        """Provides the current State UUID. This is a unique identifier
        for the current state of the cluster and the sequence of changes
        it undergoes.
        """
        return self._status('wsrep_cluster_state_uuid')

    @property
    def wsrep_cluster_status(self):
        """Status of this cluster component. That is, whether the node is
        part of a ``PRIMARY`` or ``NON_PRIMARY`` component."""
        return self._status('wsrep_cluster_status')

    @property
    def wsrep_local_state(self):
        """Internal Galera Cluster FSM state number."""
        return int(self._status('wsrep_local_state'))

    @property
    def wsrep_cluster_name(self):
        """The logical cluster name for the node."""
        result = self.execute("SELECT @@wsrep_cluster_name")
        return result[0]['@@wsrep_cluster_name']

    def execute(self, query, *args):
        """Execute query in Galera Node.

        :param query: Query to execute.
        :type query: str
        :return: Query result or None if the query is not supposed
            to return result.
        :rtype: dict
        """
        with self._connect() as conn:
            return execute(conn, query, *args)

    @contextmanager
    def _connect(self):
        """Connect to Galera node

        :return: MySQL connection to the Galera node
        :rtype: Connection
        """
        connection = pymysql.connect(  # pylint: disable=duplicate-code
            host=self.host,
            port=self.port,
            user=self.user,
            passwd=self.password,
            cursorclass=DictCursor
        )
        yield connection
        connection.close()

    def _status(self, status_variable):
        """Return value of a variable from SHOW GLOBAL STATUS"""
        result = self.execute('SHOW GLOBAL STATUS LIKE %s', status_variable)
        return result[0]['Value']

    def __eq__(self, other):
        return self.host == other.host and self.port == self.port

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "%s:%d" % (self.host, self.port)



import pytest
from pymysql import OperationalError

from proxysql_tools.proxysql.proxysql import ProxySQLMySQLBackend


def test_proxysql_ping(proxysql):
    assert proxysql.ping()


def test_execute_select(proxysql):
    assert proxysql.execute('SELECT 1') == [{'1': '1'}]
    assert proxysql.execute('SHOW TABLES') == [
        {u'tables': 'global_variables'},
        {u'tables': 'mysql_collations'},
        {u'tables': 'mysql_query_rules'},
        {u'tables': 'mysql_replication_hostgroups'},
        {u'tables': 'mysql_servers'},
        {u'tables': 'mysql_users'},
        {u'tables': 'runtime_global_variables'},
        {u'tables': 'runtime_mysql_query_rules'},
        {u'tables': 'runtime_mysql_replication_hostgroups'},
        {u'tables': 'runtime_mysql_servers'},
        {u'tables': 'runtime_mysql_users'},
        {u'tables': 'runtime_scheduler'},
        {u'tables': 'scheduler'}
    ]


def test_execute_set(proxysql):
    assert proxysql.execute('set mysql-wait_timeout=28800000') == ()


def test_execute_wrong_sql(proxysql):
    with pytest.raises(OperationalError):
        assert proxysql.execute('aaa')


def test_register_backend(proxysql):
    backend = ProxySQLMySQLBackend('192.168.90.2', hostgroup_id=10)
    proxysql.register_backend(backend)
    proxysql.deregister_backend(backend)


def test_writer(proxysql):
    backend = ProxySQLMySQLBackend('192.168.90.2', hostgroup_id=10)
    proxysql.register_backend(backend)
    assert backend == proxysql.find_backends(10)[0]



import pytest
from mock import mock_open, mock
from pymysql import OperationalError

from proxysql_tools.util.bug1258464 import get_my_cnf, get_pid, bug1258464


def test__get_my_cnf_when_dist_not_supported(mocker):
    mock_func = mocker.patch('proxysql_tools.util.bug1258464.platform').linux_distribution
    mock_func.return_value = ['0']
    with pytest.raises(NotImplementedError):
        get_my_cnf()


def test__get_my_cnf_when_dist_is_deb_based(mocker):
    mock_func = mocker.patch('proxysql_tools.util.bug1258464.platform').linux_distribution
    mock_func.return_value = ['debian']
    assert get_my_cnf() == '/etc/mysql/my.cnf'


def test__get_my_cnf_when_dist_is_rhel_based(mocker):
    mock_func = mocker.patch('proxysql_tools.util.bug1258464.platform').linux_distribution
    mock_func.return_value = ['RHEL']
    assert get_my_cnf() == '/etc/my.cnf'


def test__get_pid_returns_integer_value(mocker):
    mock_func = mocker.patch('proxysql_tools.util.bug1258464.ConfigParser').get
    mock_func.return_value = 'some path'
    mocker.patch('proxysql_tools.util.bug1258464.open', mock_open(read_data='3333'))
    assert type(get_pid('foo-bar')) is int

def test__bug1258464_when_condition_for_kill_is_true(mocker):
    mock_pymysql = mocker.patch('proxysql_tools.util.bug1258464.pymysql')
    mock_cursor = mock.MagicMock()
    mock_cursor.fetchone.return_value = [101]
    mock_pymysql.connect.return_value.cursor.return_value.__enter__.return_value = mock_cursor
    mock_get_my_cnf = mocker.patch('proxysql_tools.util.bug1258464.get_my_cnf')
    mock_get_my_cnf.return_value = 'some path'
    mock_get_pid = mocker.patch('proxysql_tools.util.bug1258464.get_pid')
    mock_get_pid.return_value = 300
    mock_kill_process = mocker.patch('proxysql_tools.util.bug1258464.kill_process')
    assert bug1258464('some path')


def test__bug1258464_when_condition_for_kill_is_false(mocker):
    mock_pymysql = mocker.patch('proxysql_tools.util.bug1258464.pymysql')
    mock_cursor = mock.MagicMock()
    mock_cursor.fetchone.return_value = [100]
    mock_pymysql.connect.return_value.cursor.return_value.__enter__.return_value = mock_cursor
    mock_get_my_cnf = mocker.patch('proxysql_tools.util.bug1258464.get_my_cnf')
    mock_get_my_cnf.return_value = 'some path'
    mock_get_pid = mocker.patch('proxysql_tools.util.bug1258464.get_pid')
    mock_get_pid.return_value = 300
    mock_kill_process = mocker.patch('proxysql_tools.util.bug1258464.kill_process')
    assert not bug1258464('some path')


def test__bug1258464_when_raise_NotImplementedError(mocker):
    mock_pymysql = mocker.patch('proxysql_tools.util.bug1258464.pymysql')
    mock_cursor = mock.MagicMock()
    mock_cursor.fetchone.return_value = [100]
    mock_pymysql.connect.return_value.cursor.return_value.__enter__.return_value = mock_cursor
    mock_get_my_cnf = mocker.patch('proxysql_tools.util.bug1258464.get_my_cnf')
    mock_get_my_cnf.return_value = 'some path'
    mock_get_my_cnf.side_effect = NotImplementedError
    assert not bug1258464('some path')


def test__bug1258464_when_raise_OSError(mocker):
    mock_pymysql = mocker.patch('proxysql_tools.util.bug1258464.pymysql')
    mock_cursor = mock.MagicMock()
    mock_cursor.fetchone.return_value = [100]
    mock_pymysql.connect.return_value.cursor.return_value.__enter__.return_value = mock_cursor
    mock_get_my_cnf = mocker.patch('proxysql_tools.util.bug1258464.get_my_cnf')
    mock_get_my_cnf.return_value = 'some path'
    mock_get_pid = mocker.patch('proxysql_tools.util.bug1258464.get_pid')
    mock_get_pid.side_effect = OSError
    assert not bug1258464('some path')


def test__bug1258464_when_raise_OperationalError(mocker):
    mock_pymysql = mocker.patch('proxysql_tools.util.bug1258464.pymysql')
    mock_cursor = mock.MagicMock()
    mock_cursor.fetchone.return_value = [100]
    mock_pymysql.connect.return_value.cursor.return_value.__enter__.return_value = mock_cursor
    mock_pymysql.connect.side_effect = OperationalError
    assert not bug1258464('some path')



"""
Module with functions to workaround bug
https://bugs.launchpad.net/percona-xtradb-cluster/+bug/1258464
"""
import os
import platform
import signal
from ConfigParser import ConfigParser
import pymysql
from pymysql import OperationalError

from proxysql_tools import LOG


def kill_process(pid):
    """Kill UNIX process.

    :param pid: Process ID.
    :type pid: int
    """
    os.kill(pid, signal.SIGKILL)


def get_pid(path):
    """Read MySQL PID from config file.

    :param path: Path to proxysql_tools config file.
    :type path: str
    """
    parser_my_cnf = ConfigParser(allow_no_value=True)
    parser_my_cnf.read(path)
    pid_file = parser_my_cnf.get('mysqld', 'pid-file')
    with open(pid_file, 'r') as file_desc:
        pid = file_desc.readline()
        return int(pid)


def get_my_cnf():
    """Detect Operating System type and return path to my.cnf.

    :return: Path to my.cnf typical for detected OS.
    :raise: NotImplementedError if failed to detect OS."""
    dist_name = platform.linux_distribution()[0]
    if dist_name.upper() in ["DEBIAN", "UBUNTU"]:
        return '/etc/mysql/my.cnf'
    elif dist_name.upper() in ["RHEL", "CENTOS", "FEDORA"]:
        return '/etc/my.cnf'
    else:
        raise NotImplementedError


def bug1258464(default_file):
    """Check if node is affected by
    https://bugs.launchpad.net/percona-xtradb-cluster/+bug/1258464
    and if yes, kill the node.

    :param default_file: Path to my.cnf
    :return: True if it killed the node.
    :rtype: bool"""
    try:
        connection = pymysql.connect(read_default_file=default_file)
        with connection.cursor() as cursor:
            cursor.execute("SELECT COUNT(*)"
                           "FROM information_schema.processlist "
                           "WHERE State = 'wsrep in pre-commit stage' "
                           "AND Info = 'COMMIT'")
            count_pre_commit = cursor.fetchone()[0]
            cursor.execute("SELECT COUNT(*)"
                           "FROM information_schema.processlist "
                           "WHERE State = "
                           "'Waiting for table metadata lock' "
                           "AND Info LIKE 'ALTER TABLE%'")
            count_waiting = cursor.fetchone()[0]
            if count_pre_commit > 100 and count_waiting > 0:
                my_cnf_path = get_my_cnf()
                pid = get_pid(my_cnf_path)
                kill_process(pid)
                return True
    except OperationalError as err:
        LOG.error(str(err))
    except OSError as err:
        LOG.error(str(err))
    except NotImplementedError as err:
        LOG.error(str(err))
    return False



import pytest

from proxysql_tools import setup_logging, LOG
from proxysql_tools.galera.galera_cluster import GaleraCluster
from proxysql_tools.galera.galera_node import GaleraNode
from proxysql_tools.proxysql.proxysql import ProxySQL

setup_logging(LOG, debug=True)


@pytest.fixture
def galera_node1():
    return GaleraNode('192.168.90.2', password='root')


@pytest.fixture
def galera_cluster():
    hosts = "192.168.90.2:3306,192.168.90.3:3306,192.168.90.4:3306"
    return GaleraCluster(hosts, password='root')


@pytest.fixture
def proxysql():
    return ProxySQL(host='127.0.0.1', port=6032,
                    user='admin', password='admin')



"""Module implements load balancing algorithms"""
from pymysql import OperationalError

from proxysql_tools import LOG
from proxysql_tools.galera.exceptions import GaleraClusterSyncedNodeNotFound, \
    GaleraClusterNodeNotFound
from proxysql_tools.galera.galera_node import GaleraNodeState, GaleraNode
from proxysql_tools.proxysql.exceptions import ProxySQLBackendNotFound
from proxysql_tools.proxysql.proxysql import ProxySQLMySQLBackend, BackendStatus


def singlewriter(galera_cluster, proxysql,
                 writer_hostgroup_id,
                 reader_hostgroup_id,
                 ignore_writer=None):
    """
    Implements singlewriter balancing mode.

    :param galera_cluster: GaleraCluster instance.
    :type galera_cluster: GaleraCluster
    :param proxysql: ProxySQL instance
    :type proxysql: ProxySQL
    :param writer_hostgroup_id: Writer hostgroup_id
    :type writer_hostgroup_id: int
    :param reader_hostgroup_id: Reader hostgroup_id
    :type reader_hostgroup_id: int
    :param ignore_writer: Do not make this backend writer
    :type ignore_writer: ProxySQLMySQLBackend
    """
    register_writer(galera_cluster, proxysql, writer_hostgroup_id,
                    reader_hostgroup_id,
                    ignore_writer=ignore_writer)
    register_readers(galera_cluster, proxysql, writer_hostgroup_id,
                     reader_hostgroup_id)

    LOG.debug('Register all missing backends')
    for galera_node in galera_cluster.find_synced_nodes():
        reader = ProxySQLMySQLBackend(galera_node.host,
                                      hostgroup_id=reader_hostgroup_id,
                                      port=galera_node.port,
                                      comment='Reader')
        writer = ProxySQLMySQLBackend(galera_node.host,
                                      hostgroup_id=writer_hostgroup_id,
                                      port=galera_node.port)
        if not (proxysql.backend_registered(reader) or
                proxysql.backend_registered(writer)):
            proxysql.register_backend(reader)
            LOG.info('Added backend %s to hostgroup %d', reader,
                     reader_hostgroup_id)

    LOG.debug('Make sure writer is not reader')
    writer = proxysql.find_backends(writer_hostgroup_id)[0]
    readers = proxysql.find_backends(reader_hostgroup_id)
    writer_as_reader = writer
    writer_as_reader.hostgroup_id = reader_hostgroup_id

    is_readers_offline = False
    if writer_as_reader in readers:
        readers_without_writer = readers[:]
        readers_without_writer.remove(writer_as_reader)

        is_readers_offline = all(x.status == BackendStatus.offline_soft
                                 for x in readers_without_writer)

    if len(readers) > 2 and proxysql.backend_registered(writer_as_reader) \
        and not is_readers_offline:
        proxysql.deregister_backend(writer_as_reader)


def register_writer(galera_cluster, proxysql, writer_hostgroup_id,
                    reader_hostgroup_id,
                    ignore_writer=None):
    """
    Checks ProxySQL and Galera cluster and makes sure there is one
    healthy registered writer.

    :param galera_cluster: GaleraCluster instance.
    :type galera_cluster: GaleraCluster
    :param proxysql: ProxySQL instance
    :type proxysql: ProxySQL
    :param writer_hostgroup_id: Writer hostgroup_id
    :type writer_hostgroup_id: int
    :param reader_hostgroup_id: Reader hostgroup_id
    :type reader_hostgroup_id: int
    :param ignore_writer: Do not make this backend writer
    :type ignore_writer: ProxySQLMySQLBackend
    """
    # Get writer backend. If doesn't exist - add.
    # If exists - check state of respective Galera node.
    # if the Galera node is not healthy - delete the backend and add a healthy
    # one.
    LOG.debug('Registering writers')
    try:
        for backend in proxysql.find_backends(writer_hostgroup_id):
            check_backend(backend, galera_cluster, proxysql,
                          writer_hostgroup_id, 'Writer', limit=1,
                          ignore_backend=ignore_writer,
                          recovered_hostgroup_id=reader_hostgroup_id,
                          recoverd_comment='Reader')

    except ProxySQLBackendNotFound:
        # add it
        register_synced_backends(galera_cluster, proxysql,
                                 writer_hostgroup_id,
                                 comment='Writer',
                                 limit=1,
                                 ignore_backend=ignore_writer)

    try:
        proxysql.find_backends(writer_hostgroup_id, BackendStatus.online)
    except ProxySQLBackendNotFound:
        LOG.warn('No writer backends were registered. '
                 'Will try to add previously ignored backends')
        register_synced_backends(galera_cluster, proxysql,
                                 writer_hostgroup_id,
                                 comment='Writer',
                                 limit=1)



def register_readers(galera_cluster, proxysql,
                     writer_hostgroup_id, reader_hostgroup_id,
                     ignore_writer=None):
    """
    Checks ProxySQL and Galera cluster and makes sure readers are registered.

    :param galera_cluster: GaleraCluster instance.
    :type galera_cluster: GaleraCluster
    :param proxysql: ProxySQL instance
    :type proxysql: ProxySQL
    :param writer_hostgroup_id: Writer hostgroup_id
    :type writer_hostgroup_id: int
    :param reader_hostgroup_id: Reader hostgroup_id
    :type reader_hostgroup_id: int
    :param ignore_writer: Do not make this backend writer
    :type ignore_writer: ProxySQLMySQLBackend
    """
    LOG.debug('Registering readers')
    try:
        writer = proxysql.find_backends(writer_hostgroup_id)[0]
    except ProxySQLBackendNotFound as err:
        LOG.warn(err)
        register_synced_backends(galera_cluster, proxysql,
                                 writer_hostgroup_id,
                                 comment='Writer',
                                 limit=1,
                                 ignore_backend=ignore_writer)
        writer = proxysql.find_backends(writer_hostgroup_id)[0]

    try:
        num_readers = 0
        readers = proxysql.find_backends(reader_hostgroup_id)
        for backend in readers:
            LOG.debug('Comparing %s and %s', backend, writer)
            if backend == writer:
                # Don't register writer as reader for now
                continue
            LOG.debug('Do not match')

            check_backend(backend, galera_cluster, proxysql,
                          reader_hostgroup_id, 'Reader')
            num_readers += 1
        if num_readers == 0:
            LOG.warn('Did not register any readers')
            check_backend(writer, galera_cluster, proxysql,
                          reader_hostgroup_id, 'Reader')

    except ProxySQLBackendNotFound:
        # If there are no readers , register writer as reader as well
        LOG.debug('Reader backends not found')
        writer_as_reader = writer
        writer_as_reader.hostgroup_id = reader_hostgroup_id
        register_synced_backends(galera_cluster, proxysql,
                                 reader_hostgroup_id,
                                 comment='Reader',
                                 ignore_backend=writer_as_reader)


def check_backend(backend, galera_cluster, proxysql, hostgroup_id, comment,  # pylint: disable=too-many-arguments
                  limit=None, ignore_backend=None,
                  recovered_hostgroup_id=None, recoverd_comment=None):
    """
    Check health of given backed and if necessary replace it.

    :param backend: MySQL backend.
    :type backend: ProxySQLMySQLBackend
    :param galera_cluster: GaleraCluster instance.
    :type galera_cluster: GaleraCluster
    :param proxysql: ProxySQL instance.
    :type proxysql: ProxySQL
    :param hostgroup_id: Hostgroup_id which the backend belongs to.
    :type hostgroup_id: int
    :param comment: Comment to add to mysql_servers in ProxySQL
    :param limit: Register not more than limit number of backends
    :type limit: int
    :param ignore_backend: Do not register this backend
    :type ignore_backend: ProxySQLMySQLBackend
    :param recovered_hostgroup_id: If backend recovers from OFFLINE_SOFT assign
        it to this hostgroup_id. Default hostgroup_id.
    :type recovered_hostgroup_id: int
    :param recoverd_comment: If backend recovers from OFFLINE_SOFT set
        this comment
    :type recoverd_comment: str
    :return: True if backend successfully registered.
    :rtype: bool
    """
    # check it
    LOG.debug('Backend %s is already registered', backend)
    LOG.debug('Checking its health')
    try:
        node = galera_cluster.find_node(backend.hostname, backend.port)

        state = node.wsrep_local_state
        LOG.debug('%s state: %d', node, state)

        if state == GaleraNodeState.SYNCED:
            LOG.debug('Node %s (%s) is healthy', node, backend.status)

            if backend.status != BackendStatus.online:

                LOG.debug('Deregistering %s (%s)', backend, backend.status)
                proxysql.deregister_backend(backend)

                backend.status = BackendStatus.online

                if not recovered_hostgroup_id:
                    recovered_hostgroup_id = hostgroup_id
                backend.hostgroup_id = recovered_hostgroup_id

                if recoverd_comment:
                    backend.comment = recoverd_comment

                LOG.debug('Registering %s (%s)', backend, backend.status)
                proxysql.register_backend(backend)
        else:
            LOG.warn('Node %s is reachable but unhealty, '
                     'setting it OFFLINE_SOFT', node)
            proxysql.set_status(backend, BackendStatus.offline_soft)
            register_synced_backends(galera_cluster, proxysql,
                                     hostgroup_id,
                                     comment=comment,
                                     limit=limit,
                                     ignore_backend=ignore_backend)

    except GaleraClusterNodeNotFound:
        LOG.warn('Backend %s is not a cluster member. Deregistering it.',
                 backend)
        proxysql.deregister_backend(backend)
        register_synced_backends(galera_cluster, proxysql,
                                 hostgroup_id, comment=comment,
                                 limit=limit, ignore_backend=ignore_backend)
    except OperationalError as err:
        LOG.error(err)
        LOG.error('Looks like backend %s is unhealthy. Deregistering it.',
                  backend)
        proxysql.deregister_backend(backend)
        register_synced_backends(galera_cluster, proxysql,
                                 hostgroup_id, comment=comment,
                                 limit=limit, ignore_backend=ignore_backend)
    return True


def register_synced_backends(galera_cluster, proxysql,  # pylint: disable=too-many-arguments
                             hostgroup_id, comment=None, limit=None,
                             ignore_backend=None):
    """
    Find SYNCED node and register it as a backend.

    :param galera_cluster: GaleraCluster instance.
    :type galera_cluster: GaleraCluster
    :param proxysql: ProxySQL instance
    :type proxysql: ProxySQL
    :param hostgroup_id: hostgroup_id
    :type hostgroup_id: int
    :param comment: Optional comment to add to mysql_server
    :type comment: str
    :param limit: Register not more than limit number of backends
    :type limit: int
    :param ignore_backend: Do not register this backend
    :type ignore_backend: ProxySQLMySQLBackend
    """
    try:
        galera_nodes = galera_cluster.find_synced_nodes()

        if ignore_backend:
            node = GaleraNode(ignore_backend.hostname,
                              port=ignore_backend.port)
            LOG.debug('Ignoring backend %s', ignore_backend)
            if node in galera_nodes:
                LOG.debug('Remove %s from candidates', ignore_backend)
                galera_nodes.remove(node)

        if limit:
            candidate_nodes = galera_nodes[:limit]
        else:
            candidate_nodes = galera_nodes

        for galera_node in candidate_nodes:
            backend = ProxySQLMySQLBackend(galera_node.host,
                                           hostgroup_id=hostgroup_id,
                                           port=galera_node.port,
                                           comment=comment)
            proxysql.register_backend(backend)
            LOG.info('Added backend %s to hostgroup %d', backend, hostgroup_id)

    except GaleraClusterSyncedNodeNotFound as err:
        LOG.error(err)



from mock import MagicMock

from proxysql_tools.aws.aws import get_network_interface, \
    get_network_interface_state, network_interface_attached, \
    detach_network_interface


def test__get_network_interface(mocker):
    mock_boto3 = mocker.patch('proxysql_tools.aws.aws.boto3')
    mock_ec2 = MagicMock()
    mock_boto3.client.return_value = mock_ec2
    get_network_interface('0.0.0.0')
    mock_boto3.client.assert_called_once_with('ec2')


def test__get_network_interface_state(mocker):
    mock_boto3 = mocker.patch('proxysql_tools.aws.aws.boto3')
    mock_ec2 = MagicMock()
    mock_boto3.client.return_value = mock_ec2
    get_network_interface_state('some interface')
    mock_boto3.client.assert_called_once_with('ec2')


def test__network_interface_state_returns_true_if_attached(mocker):
    mock_get_interface_state = mocker.patch('proxysql_tools.aws.aws.get_network_interface_state')
    mock_get_interface_state.return_value = 'attached'
    assert network_interface_attached('1.1.1.1')


def test__network_interface_state_returns_false_if_not_attached(mocker):
    mock_get_interface_state = mocker.patch('proxysql_tools.aws.aws.get_network_interface_state')
    mock_get_interface_state.return_value = 'detached'
    assert not network_interface_attached('1.1.1.1')


def test__network_interface_attached_if_side_effect_is_key_error(mocker):
    mock_get_interface_state = mocker.patch('proxysql_tools.aws.aws.get_network_interface_state')
    mock_get_interface_state.side_effect = KeyError
    assert not network_interface_attached('1.1.1.1')


def test__detach_network_interface(mocker):
    mock_boto3 = mocker.patch('proxysql_tools.aws.aws.boto3')
    mock_ec2 = MagicMock()
    mock_boto3.client.return_value = mock_ec2
    detach_network_interface('stub')
    mock_boto3.client.assert_called_once_with('ec2')



# pylint: skip-file
import os
import time
from ConfigParser import NoOptionError
from subprocess import Popen

import netifaces
import requests
import boto3
from botocore.exceptions import ClientError

from proxysql_tools import LOG

DEVICE_INDEX = 1


def get_my_instance_id():
    r = requests.get('http://169.254.169.254/latest/meta-data/instance-id')
    r.raise_for_status()
    return r.content


def get_network_interface(ip):
    client = boto3.client('ec2')
    response = client.describe_network_interfaces(
        Filters=[
            {
                'Name': 'addresses.private-ip-address',
                'Values': [
                    ip,
                ]
            },
        ]
    )
    network_interface_id = \
        response['NetworkInterfaces'][0]['NetworkInterfaceId']
    return network_interface_id


def ensure_local_interface_is_gone(local_interface):
    while local_interface in netifaces.interfaces():
        pass


def get_network_interface_state(network_interface):
    client = boto3.client('ec2')
    response = client.describe_network_interfaces(
        NetworkInterfaceIds=[network_interface]
    )
    status = response['NetworkInterfaces'][0]['Attachment']['Status']
    LOG.debug('Interface %s, status = %s', network_interface, status)
    return status


def network_interface_attached(network_interface):
    """
    Check whether network interface is attached

    :param network_interface: network interface id
    :return: True or False
    """
    try:
        status = get_network_interface_state(network_interface)
        return status == 'attached'
    except KeyError:
        return False


def detach_network_interface(network_interface):

    def get_attachment_id(boto_client, interface):
        response = boto_client.describe_network_interfaces(
            NetworkInterfaceIds=[
                interface
            ]
        )
        return response['NetworkInterfaces'][0]['Attachment']['AttachmentId']

    client = boto3.client('ec2')
    client.detach_network_interface(
        AttachmentId=get_attachment_id(client, network_interface)
    )


def ensure_network_interface_is_detached(network_interface):
    try:
        while get_network_interface_state(network_interface) != 'detached':
            time.sleep(1)
    except KeyError:
        pass


def attach_network_interface(network_interface, instance_id):
    client = boto3.client('ec2')
    for _ in xrange(10):
        try:
            client.attach_network_interface(
                NetworkInterfaceId=network_interface,
                InstanceId=instance_id,
                DeviceIndex=DEVICE_INDEX
            )
        except ClientError:
            time.sleep(1)


def configure_local_interface(local_interface, ip, netmask):

    cmd = [
        'ifconfig',
        local_interface,
        'inet',
        ip,
        'netmask',
        netmask
    ]
    for _ in xrange(10):
        os.environ['PATH'] = "%s:/sbin" % os.environ['PATH']
        env = os.environ
        proc = Popen(cmd, env=env)
        proc.communicate()
        if proc.returncode:
            time.sleep(1)
        else:
            return


def aws_notify_master(cfg):
    """The function moves network interface to local instance and brings it up.
    Steps:

    - Detach network interface if attached to anywhere.
    - Attach the network interface to the local instance.
    - Configure IP address on this instance

    :param cfg: config object
    """
    try:
        os.environ['AWS_ACCESS_KEY_ID'] = cfg.get('aws', 'aws_access_key_id')
        os.environ['AWS_SECRET_ACCESS_KEY'] = cfg.get('aws',
                                                      'aws_secret_access_key')
        os.environ['AWS_DEFAULT_REGION'] = cfg.get('aws', 'aws_default_region')
    except NoOptionError:
        LOG.error('aws_access_key_id, aws_secret_access_key and '
                  'aws_default_region must be defined in '
                  'aws section of the config file.')
        exit(-1)

    instance_id = get_my_instance_id()
    try:
        ip = cfg.get('proxysql', 'virtual_ip')
        netmask = cfg.get('proxysql', 'virtual_netmask')

        network_interface = get_network_interface(ip)

        if network_interface_attached(network_interface):
            detach_network_interface(network_interface)

        local_interface = "eth%d" % DEVICE_INDEX
        ensure_local_interface_is_gone(local_interface)

        ensure_network_interface_is_detached(network_interface)

        attach_network_interface(network_interface, instance_id)

        configure_local_interface(local_interface, ip, netmask)
    except NoOptionError as err:
        LOG.error('virtual_ip and virtual_netmask must be defined in '
                  'proxysql section of the config file.')
        LOG.error(err)



"""ProxySQL classes"""
from contextlib import contextmanager

import pymysql
from pymysql.cursors import DictCursor

from proxysql_tools import LOG, execute
from proxysql_tools.proxysql.exceptions import ProxySQLBackendNotFound

PROXYSQL_CONNECT_TIMEOUT = 20


class BackendStatus(object):  # pylint: disable=too-few-public-methods
    """Status of ProxySQL backend"""
    online = 'ONLINE'
    shunned = 'SHUNNED'
    offline_soft = 'OFFLINE_SOFT'
    offline_hard = 'OFFLINE_HARD'


class ProxySQLMySQLBackend(object):  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """ProxySQLMySQLBackend describes record in ProxySQL
    table ``mysql_servers``.

.. code-block:: mysql

    CREATE TABLE mysql_servers (
        hostgroup_id INT NOT NULL DEFAULT 0,
        hostname VARCHAR NOT NULL,
        port INT NOT NULL DEFAULT 3306,
        status VARCHAR CHECK (UPPER(status) IN
            ('ONLINE','SHUNNED','OFFLINE_SOFT', 'OFFLINE_HARD'))
            NOT NULL DEFAULT 'ONLINE',
        weight INT CHECK (weight >= 0) NOT NULL DEFAULT 1,
        compression INT CHECK (compression >=0 AND compression <= 102400)
            NOT NULL DEFAULT 0,
        max_connections INT CHECK (max_connections >=0) NOT NULL DEFAULT 1000,
        max_replication_lag INT CHECK (max_replication_lag >= 0
            AND max_replication_lag <= 126144000) NOT NULL DEFAULT 0,
        use_ssl INT CHECK (use_ssl IN(0,1)) NOT NULL DEFAULT 0,
        max_latency_ms INT UNSIGNED CHECK (max_latency_ms>=0)
            NOT NULL DEFAULT 0,
        comment VARCHAR NOT NULL DEFAULT '',
        PRIMARY KEY (hostgroup_id, hostname, port) )

    """
    def __init__(self, hostname, hostgroup_id=0, port=3306,  # pylint: disable=too-many-arguments
                 status=BackendStatus.online,
                 weight=1, compression=0, max_connections=10000,
                 max_replication_lag=0, use_ssl=False,
                 max_latency_ms=0, comment=None):
        self.hostname = hostname
        self.hostgroup_id = int(hostgroup_id)
        self.port = int(port)
        self.status = status
        self.weight = int(weight)
        self.compression = int(compression)
        self.max_connections = int(max_connections)
        self.max_replication_lag = int(max_replication_lag)
        self.use_ssl = bool(use_ssl)
        self.max_latency_ms = int(max_latency_ms)
        self.comment = comment

    def __eq__(self, other):
        try:
            return self.hostgroup_id == other.hostgroup_id and \
                   self.hostname == other.hostname and \
                   self.port == other.port
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "%d__%s__%d" % (self.hostgroup_id, self.hostname, self.port)


class ProxySQLMySQLUser(object):  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """ProxySQLMySQLUser describes record in ProxySQL table ``mysql_users``.

.. code-block:: mysql

    CREATE TABLE mysql_users (
        username VARCHAR NOT NULL,
        password VARCHAR,
        active INT CHECK (active IN (0,1)) NOT NULL DEFAULT 1,
        use_ssl INT CHECK (use_ssl IN (0,1)) NOT NULL DEFAULT 0,
        default_hostgroup INT NOT NULL DEFAULT 0,
        default_schema VARCHAR,
        schema_locked INT CHECK (schema_locked IN (0,1)) NOT NULL DEFAULT 0,
        transaction_persistent INT CHECK (transaction_persistent IN (0,1))
            NOT NULL DEFAULT 0,
        fast_forward INT CHECK (fast_forward IN (0,1)) NOT NULL DEFAULT 0,
        backend INT CHECK (backend IN (0,1)) NOT NULL DEFAULT 1,
        frontend INT CHECK (frontend IN (0,1)) NOT NULL DEFAULT 1,
        max_connections INT CHECK (max_connections >=0) NOT NULL DEFAULT 10000,
        PRIMARY KEY (username, backend),
        UNIQUE (username, frontend))


:param user: MySQL username to connect to ProxySQL or Galera node.
:param password: MySQL password.
:param active: Users with active = 0 will be tracked in the database,
    but will be never loaded in the in-memory data structures.
:type active: bool
:param use_ssl: Use SSL to connect to MySQL or not
:type use_ssl: bool
:param default_hostgroup: If there is no matching rule for the queries sent by
    the users, the traffic it generates is sent to the specified hostgroup_.
:param default_schema: The schema to which the connection should change
    by default.
:param schema_locked: not supported yet.
:type schema_locked: bool
:param transaction_persistent: if this is set for the user with which the MySQL
    client is connecting to ProxySQL (thus a "frontend" user - see below),
    transactions started within a hostgroup will remain within that hostgroup
    regardless of any other rules.
:type transaction_persistent: bool
:param fast_forward: If set, it bypasses the query processing layer (rewriting,
    caching) and passes the query directly to the backend server.
:type fast_forward: bool
:param frontend: If True, this (username, password) pair is used for
    authenticating to the ProxySQL instance.
:type frontend: bool
:param backend: If True, this (username, password) pair is used for
    authenticating to the mysqld servers against any hostgroup.
:param max_connections: Maximum number of connection this user
    can create to MySQL node.

.. _hostgroup: http://bit.ly/2rGnT5i
    """
    def __init__(self, user='root', password=None, active=False, use_ssl=False,  # pylint: disable=too-many-arguments
                 default_hostgroup=0, default_schema='information_schema',
                 schema_locked=False, transaction_persistent=False,
                 fast_forward=False, backend=False, frontend=True,
                 max_connections=10000):
        self.user = user
        self.password = password
        self.active = active
        self.use_ssl = use_ssl
        self.default_hostgroup = int(default_hostgroup)
        self.default_schema = default_schema
        self.schema_locked = schema_locked
        self.transaction_persistent = transaction_persistent
        self.fast_forward = fast_forward
        self.backend = backend
        self.frontend = frontend
        self.max_connections = int(max_connections)


class ProxySQL(object):
    """
    ProxySQL describes a single ProxySQL instance.

    :param host: ProxySQL hostname.
    :param port: Port on which ProxySQL listens to admin connections.
    :param user: ProxySQL admin user.
    :param password: Password for ProxySQL admin.
    :param socket: Socket to connect to ProxySQL admin interface.
    """
    def __init__(self, host='localhost', port=3306, user='root',  # pylint: disable=too-many-arguments
                 password=None, socket=None):

        self.host = host
        self.port = int(port)
        self.user = user
        self.password = password
        self.socket = socket

    def ping(self):
        """Check health of ProxySQL.

        :return: True if ProxySQL healthy and False otherwise.
        :rtype: bool"""
        try:
            result = self.execute('SELECT 1 AS result')
            return result[0]['result'] == '1'
        except pymysql.err.OperationalError:
            LOG.debug('ProxySQL %s:%d is dead', self.host, self.port)
            return False

    def execute(self, query, *args):
        """Execute query in ProxySQL.

        :param query: Query to execute.
        :type query: str
        :return: Query result or None if the query is not supposed
            to return result
        :rtype: dict
        """
        with self._connect() as conn:
            return execute(conn, query, *args)

    def reload_runtime(self):
        """Reload the ProxySQL runtime configuration."""
        self.execute('LOAD MYSQL SERVERS TO RUNTIME')
        self.execute('LOAD MYSQL USERS TO RUNTIME')
        self.execute('LOAD MYSQL VARIABLES TO RUNTIME')

    def register_backend(self, backend):
        """Register Galera node in ProxySQL

        :param backend: Galera node.
        :type backend: ProxySQLMySQLBackend
        """
        if backend.comment:
            comment = "'%s'" % pymysql.escape_string(backend.comment)
        else:
            comment = 'NULL'

        query = "REPLACE INTO mysql_servers(`hostgroup_id`," \
                " `hostname`, `port`," \
                " `status`, `weight`, `compression`, `max_connections`," \
                " `max_replication_lag`, `use_ssl`, `max_latency_ms`," \
                " `comment`) " \
                "VALUES({hostgroup_id}, '{hostname}', {port}," \
                " '{status}', {weight}, {compression}, {max_connections}," \
                " {max_replication_lag}, {use_ssl}, {max_latency_ms}," \
                " {comment})" \
                "".format(hostgroup_id=int(backend.hostgroup_id),
                          hostname=pymysql.escape_string(backend.hostname),
                          port=int(backend.port),
                          status=pymysql.escape_string(backend.status),
                          weight=int(backend.weight),
                          compression=int(backend.compression),
                          max_connections=int(backend.max_connections),
                          max_replication_lag=int(backend.max_replication_lag),
                          use_ssl=int(backend.use_ssl),
                          max_latency_ms=int(backend.max_latency_ms),
                          comment=comment)
        self.execute(query)
        self.reload_runtime()

    def deregister_backend(self, backend):
        """
        Deregister a Galera node from ProxySQL

        :param backend: Galera node.
        :type backend: ProxySQLMySQLBackend
        """
        query = "DELETE FROM mysql_servers WHERE hostgroup_id={hostgroup_id}" \
                " AND hostname='{hostname}'" \
                " AND port={port}" \
                "".format(hostgroup_id=int(backend.hostgroup_id),
                          hostname=pymysql.escape_string(backend.hostname),
                          port=int(backend.port))
        self.execute(query)
        self.reload_runtime()

    def find_backends(self, hostgroup_id, status=None):
        """
        Get writer from mysql_servers

        :param hostgroup_id: writer hostgroup_id
        :type hostgroup_id: int
        :param status: Look only for backends in this status
        :type status: BackendStatus
        :return: Writer MySQL backend or None if doesn't exist
        :rtype: ProxySQLMySQLBackend
        :raise: ProxySQLBackendNotFound
        """
        result = self.execute('SELECT `hostgroup_id`, `hostname`, '
                              '`port`, `status`, `weight`, `compression`, '
                              '`max_connections`, `max_replication_lag`, '
                              '`use_ssl`, `max_latency_ms`, `comment`'
                              ' FROM `mysql_servers`'
                              ' WHERE hostgroup_id = %s', hostgroup_id)

        backends = []
        for row in result:
            backend = ProxySQLMySQLBackend(row['hostname'],
                                           hostgroup_id=row['hostgroup_id'],
                                           port=row['port'],
                                           status=row['status'],
                                           weight=row['weight'],
                                           compression=row['compression'],
                                           max_connections=
                                           row['max_connections'],
                                           max_replication_lag=
                                           row['max_replication_lag'],
                                           use_ssl=row['use_ssl'],
                                           max_latency_ms=
                                           row['max_latency_ms'],
                                           comment=row['comment'])
            if status and backend.status != status:
                continue
            backends.append(backend)
        if backends:
            return backends
        else:
            raise ProxySQLBackendNotFound('Can not find any backends')

    def backend_registered(self, backend):
        """
        Check if backend is registered.

        :param backend: ProxySQLMySQLBackend instance
        :return: True if registered, False otherwise
        :rtype: bool
        """
        result = self.execute('SELECT `hostgroup_id`, `hostname`, '
                              '`port`'
                              ' FROM `mysql_servers`'
                              ' WHERE hostgroup_id = %s '
                              ' AND `hostname` = %s '
                              ' AND `port` = %s',
                              (
                                  backend.hostgroup_id,
                                  backend.hostname,
                                  backend.port
                              ))
        return result != ()

    def set_status(self, backend, status):
        """Update status of a backend in ProxySQL"""
        self.execute('UPDATE `mysql_servers` SET `status` = %s '
                     ' WHERE hostgroup_id = %s '
                     ' AND `hostname` = %s '
                     ' AND `port` = %s',
                     (
                         status,
                         backend.hostgroup_id,
                         backend.hostname,
                         backend.port
                     ))
        self.reload_runtime()

    @contextmanager
    def _connect(self):
        """Connect to ProxySQL admin interface."""
        if self.socket is not None:
            conn = pymysql.connect(unix_socket=self.socket,
                                   user=self.user,
                                   passwd=self.password,
                                   connect_timeout=PROXYSQL_CONNECT_TIMEOUT,
                                   cursorclass=DictCursor)
        else:
            conn = pymysql.connect(host=self.host, port=self.port,
                                   user=self.user, passwd=self.password,
                                   connect_timeout=PROXYSQL_CONNECT_TIMEOUT,
                                   cursorclass=DictCursor)

        yield conn
        conn.close()



"""Galera entrypoints"""
from ConfigParser import NoOptionError

from proxysql_tools import LOG
from proxysql_tools.galera.galera_cluster import GaleraCluster
from proxysql_tools.load_balancing_mode import singlewriter
from proxysql_tools.proxysql.proxysql import ProxySQL, ProxySQLMySQLBackend


def galera_register(cfg):
    """Registers Galera cluster nodes with ProxySQL."""

    kwargs = {}
    try:
        kwargs['user'] = cfg.get('galera', 'cluster_username')
    except NoOptionError:
        pass
    try:
        kwargs['password'] = cfg.get('galera', 'cluster_password')
    except NoOptionError:
        pass

    LOG.debug('Galera config %r', kwargs)
    galera_cluster = GaleraCluster(cfg.get('galera', 'cluster_host'),
                                   **kwargs)

    kwargs = {}
    try:
        kwargs['host'] = cfg.get('proxysql', 'host')
    except NoOptionError:
        pass
    try:
        kwargs['port'] = cfg.get('proxysql', 'admin_port')
    except NoOptionError:
        pass
    try:
        kwargs['user'] = cfg.get('proxysql', 'admin_username')
    except NoOptionError:
        pass
    try:
        kwargs['password'] = cfg.get('proxysql', 'admin_password')
    except NoOptionError:
        pass
    try:
        kwargs['socket'] = cfg.get('proxysql', 'admin_socket')
    except NoOptionError:
        pass
    LOG.debug('ProxySQL config %r', kwargs)
    proxysql = ProxySQL(**kwargs)

    load_balancing_mode = cfg.get('galera', 'load_balancing_mode')
    writer_hostgroup_id = int(cfg.get('galera', 'writer_hostgroup_id'))
    reader_hostgroup_id = int(cfg.get('galera', 'reader_hostgroup_id'))

    if load_balancing_mode == 'singlewriter':
        kwargs = {}
        try:
            host, port = cfg.get('galera', 'writer_blacklist').split(':')
            bcknd = ProxySQLMySQLBackend(host,
                                         hostgroup_id=writer_hostgroup_id,
                                         port=port)
            kwargs['ignore_writer'] = bcknd
        except NoOptionError:
            pass
        singlewriter(galera_cluster, proxysql,
                     writer_hostgroup_id, reader_hostgroup_id, **kwargs)
    else:
        raise NotImplementedError('Balancing mode %s not implemented yet.'
                                  % load_balancing_mode)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

from pip.req import parse_requirements
from setuptools import setup, find_packages

del os.link

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [str(ir.req) for ir in
                parse_requirements('requirements.txt', session=False)]

test_requirements = [str(ir.req) for ir in
                     parse_requirements('requirements_dev.txt', session=False)]


setup(
    name='proxysql_tools',
    version='0.3.4',
    description="ProxySQL Tools",
    long_description=readme + '\n\n' + history,
    author="TwinDB Development Team",
    author_email='dev@twindb.com',
    url='https://github.com/twindb/proxysql-tools',
    packages=find_packages(exclude=('tests*',)),
    package_dir={'proxysql_tools':
                 'proxysql_tools'},
    entry_points={
        'console_scripts': [
            'proxysql-tool=proxysql_tools.cli:main',
            'proxysql_tools=proxysql_tools.cli:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords='proxysql_tools',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



import os
import pprint
import pytest

from docker.types import IPAMConfig, IPAMPool

from proxysql_tools import LOG, setup_logging
from proxysql_tools.galera.galera_node import GaleraNode
from proxysql_tools.proxysql.proxysql import ProxySQL
from tests.integration.library import (
    docker_client,
    docker_pull_image,
    eventually,
    create_percona_xtradb_cluster
)


CONTAINERS_FOR_TESTING_LABEL = 'pytest_docker'
DEBIAN_IMAGE = 'debian:8'
PROXYSQL_IMAGE = 'twindb/proxysql:latest'
PXC_IMAGE = 'twindb/percona-xtradb-cluster:latest'

PROXYSQL_ADMIN_PORT = 6032
PROXYSQL_CLIENT_PORT = 6033
PROXYSQL_ADMIN_USER = 'admin'
PROXYSQL_ADMIN_PASSWORD = 'admin'

PXC_MYSQL_PORT = 3306
PXC_ROOT_PASSWORD = 'r00t'

setup_logging(LOG, debug=True)


def pytest_runtest_logreport(report):
    """Process a test setup/call/teardown report relating to the respective
        phase of executing a test."""
    if report.failed:
        client = docker_client()
        api = client.api

        test_containers = api.containers(
            all=True, filters={"label": CONTAINERS_FOR_TESTING_LABEL})

        for container in test_containers:
            log_lines = [
                ("docker inspect {!r}:".format(container['Id'])),
                (pprint.pformat(api.inspect_container(container['Id']))),
                ("docker logs {!r}:".format(container['Id'])),
                (api.logs(container['Id'])),
            ]
            report.longrepr.addsection('docker logs',
                                       os.linesep.join(log_lines))


@pytest.yield_fixture
def debian_container():
    client = docker_client()
    api = client.api

    # Pull the image locally
    docker_pull_image(DEBIAN_IMAGE)

    container = api.create_container(
        image=DEBIAN_IMAGE, labels=[CONTAINERS_FOR_TESTING_LABEL],
        command='/bin/sleep 36000')
    LOG.debug('Starting container %s', container['Id'])
    api.start(container['Id'])

    container_info = client.containers.get(container['Id'])

    yield container_info

    api.remove_container(container=container['Id'], force=True)


@pytest.yield_fixture(scope='session')
def container_network():
    client = docker_client()
    api = client.api

    network_name = 'test_network'

    ipam_pool = IPAMPool(subnet="172.25.0.0/16")
    ipam_config = IPAMConfig(pool_configs=[ipam_pool])

    network = api.create_network(name=network_name, driver="bridge",
                                 ipam=ipam_config)

    yield network_name

    api.remove_network(net_id=network['Id'])


@pytest.fixture
def proxysql_config_contents():
    return """
datadir="/var/lib/proxysql"

admin_variables=
{{
        admin_credentials="{admin_user}:{admin_password}"
        mysql_ifaces="0.0.0.0:{admin_port};/var/lib/proxysql/proxysql_admin.sock"
        refresh_interval=2000
        debug=false
}}

mysql_variables=
{{
        default_query_delay=0
        default_query_timeout=36000000
        have_compress=true
        poll_timeout=2000
        stacksize=1048576
        server_version="5.5.30"
        threads=8
        max_connections=2048
        interfaces="0.0.0.0:{client_port};/var/lib/proxysql/proxysql.sock"
        default_schema="information_schema"
        connect_timeout_server=1000
        connect_timeout_server_max=3000
        connect_retries_on_failure=3
        monitor_enabled=true
        monitor_history=600000
        monitor_connect_interval=60000
        monitor_ping_interval=1000
        monitor_ping_timeout=100
        ping_interval_server=10000
        ping_timeout_server=500
        commands_stats=true
        sessions_sort=true
}}""".format(admin_user=PROXYSQL_ADMIN_USER,
             admin_password=PROXYSQL_ADMIN_PASSWORD,
             admin_port=PROXYSQL_ADMIN_PORT,
             client_port=PROXYSQL_CLIENT_PORT)


@pytest.yield_fixture
def proxysql_container(proxysql_config_contents, tmpdir, container_network):
    client = docker_client()
    api = client.api

    # Pull the container image locally first
    docker_pull_image(PROXYSQL_IMAGE)

    # Setup the ProxySQL config
    config = tmpdir.join('proxysql.cnf')
    config.write(proxysql_config_contents)
    LOG.debug('ProxySQL Config:\n %s', proxysql_config_contents)

    # The ports that the ProxySQL container will be listening on inside the
    # container
    container_ports = [PROXYSQL_ADMIN_PORT, PROXYSQL_CLIENT_PORT]
    container_info = {
        'name': 'proxysql01',
        'ip': '172.25.3.100',
    }

    host_config = api.create_host_config(
        binds=["{}:/etc/proxysql.cnf".format(str(config))],
        port_bindings={}
    )

    networking_config = api.create_networking_config({
        container_network: api.create_endpoint_config(
            ipv4_address=container_info['ip']
        )
    })

    container = api.create_container(image=PROXYSQL_IMAGE,
                                     name='proxysql01',
                                     labels=[CONTAINERS_FOR_TESTING_LABEL],
                                     ports=container_ports,
                                     host_config=host_config,
                                     networking_config=networking_config)
    LOG.debug('Starting container %s', container['Id'])
    api.start(container['Id'])
    # 1/0

    yield container_info

    api.remove_container(container=container['Id'], force=True)


@pytest.fixture
def proxysql_instance(proxysql_container):
    LOG.debug("Container %r", proxysql_container)
    connection = ProxySQL(host=proxysql_container['ip'],
                          port=PROXYSQL_ADMIN_PORT,
                          user=PROXYSQL_ADMIN_USER,
                          password=PROXYSQL_ADMIN_PASSWORD)

    def check_started():
        LOG.debug('Checking if proxysql is up')
        ret = connection.ping()
        LOG.debug(ret)
        return ret

    # Allow ProxySQL to startup completely. The problem is that ProxySQL starts
    # listening to the admin port before it has initialized completely which
    # causes the test to fail with the exception:
    # OperationalError: (2013, 'Lost connection to MySQL server during query')
    eventually(check_started, retries=15, sleep_time=4)

    return connection


@pytest.yield_fixture
def percona_xtradb_cluster_three_node(container_network):
    client = docker_client()
    api = client.api

    # The ports that need to be exposed from the PXC container to host
    container_ports = [PXC_MYSQL_PORT]
    container_info = [
        {
            'id': None,
            'name': 'pxc-node01',
            'ip': '172.25.3.1',
            'mysql_port': PXC_MYSQL_PORT,
            'root_password': PXC_ROOT_PASSWORD
        },
        {
            'id': None,
            'name': 'pxc-node02',
            'ip': '172.25.3.2',
            'mysql_port': PXC_MYSQL_PORT,
            'root_password': PXC_ROOT_PASSWORD
        },
        {
            'id': None,
            'name': 'pxc-node03',
            'ip': '172.25.3.3',
            'mysql_port': PXC_MYSQL_PORT,
            'root_password': PXC_ROOT_PASSWORD
        }
    ]
    cluster_name = 'test_cluster_3_node'

    # Create the cluster
    container_info = create_percona_xtradb_cluster(
        PXC_IMAGE, [CONTAINERS_FOR_TESTING_LABEL], container_info,
        container_ports, container_network, cluster_name
    )

    yield container_info

    # Cleanup the containers now
    for container in container_info:
        api.remove_container(container=container['id'], force=True)


@pytest.yield_fixture
def percona_xtradb_cluster_one_node(container_network):
    client = docker_client()
    api = client.api

    # The ports that need to be exposed from the PXC container to host
    container_ports = [PXC_MYSQL_PORT]
    container_info = [
        {
            'id': None,
            'name': 'pxc-one-node-node01',
            'ip': '172.25.3.10',
            'mysql_port': PXC_MYSQL_PORT,
            'root_password': PXC_ROOT_PASSWORD
        }
    ]
    cluster_name = 'test_cluster_1_node'

    # Create the cluster
    container_info = create_percona_xtradb_cluster(
        PXC_IMAGE, [CONTAINERS_FOR_TESTING_LABEL], container_info,
        container_ports, container_network, cluster_name
    )

    yield container_info

    # Cleanup the containers now
    for container in container_info:
        api.remove_container(container=container['id'], force=True)


@pytest.fixture
def percona_xtradb_cluster_node(percona_xtradb_cluster_one_node):
    node = GaleraNode(host=percona_xtradb_cluster_one_node[0]['ip'],
                      user='root',
                      password=PXC_ROOT_PASSWORD
    )

    def check_started():
        node.execute('SELECT 1')
        return True

    # Allow the cluster node to startup completely.
    eventually(check_started, retries=15, sleep_time=4)

    return node



#import pymysql
#from pymysql.cursors import DictCursor
#from proxysql_tools.proxysql.proxysql import (
#    ProxySQLMySQLBackend,
#)
#
#
#def test__register_cluster_with_proxysql_is_idempotent(
#        percona_xtradb_cluster_node, proxysql_instance):
#    hostgroup_writer = 10
#    hostgroup_reader = 11
#
#
#    mysql_backend_writer = ProxySQLMySQLBackend(hostname=percona_xtradb_cluster_node.host,
#                                         port=percona_xtradb_cluster_node.port,
#                                         hostgroup_id=hostgroup_writer
#    )
#    proxysql_instance.register_backend(mysql_backend_writer)
#
#    mysql_backend_reader = ProxySQLMySQLBackend(hostname=percona_xtradb_cluster_node.host,
#                                         port=percona_xtradb_cluster_node.port,
#                                         hostgroup_id=hostgroup_reader
#    )
#    proxysql_instance.register_backend(mysql_backend_reader)
#
#
#    connection = pymysql.connect(host=proxysql_instance.host, port=proxysql_instance.port,
#                           user=proxysql_instance.user, passwd=proxysql_instance.password,
#                           connect_timeout=20,
#                           cursorclass=DictCursor)
#
#    try:
#        with connection.cursor() as cursor:
#            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
#                           '`port`, `status`, `weight`, `compression`, '
#                           '`max_connections`, `max_replication_lag`, '
#                           '`use_ssl`, `max_latency_ms`, `comment`'
#                           ' FROM `mysql_servers`'
#                           ' WHERE hostgroup_id = %s', hostgroup_writer)
#            row = cursor.fetchall()[0]
#            backend_writer = ProxySQLMySQLBackend(row['hostname'],
#                                           hostgroup_id=row['hostgroup_id'],
#                                           port=row['port'],
#                                           status=row['status'],
#                                           weight=row['weight'],
#                                           compression=row['compression'],
#                                           max_connections=
#                                           row['max_connections'],
#                                           max_replication_lag=
#                                           row['max_replication_lag'],
#                                           use_ssl=row['use_ssl'],
#                                           max_latency_ms=
#                                           row['max_latency_ms'],
#                                           comment=row['comment'])
#            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
#                           '`port`, `status`, `weight`, `compression`, '
#                           '`max_connections`, `max_replication_lag`, '
#                           '`use_ssl`, `max_latency_ms`, `comment`'
#                           ' FROM `mysql_servers`'
#                           ' WHERE hostgroup_id = %s', hostgroup_reader)
#            row = cursor.fetchall()[0]
#            backend_reader = ProxySQLMySQLBackend(row['hostname'],
#                                           hostgroup_id=row['hostgroup_id'],
#                                           port=row['port'],
#                                           status=row['status'],
#                                           weight=row['weight'],
#                                           compression=row['compression'],
#                                           max_connections=
#                                           row['max_connections'],
#                                           max_replication_lag=
#                                           row['max_replication_lag'],
#                                           use_ssl=row['use_ssl'],
#                                           max_latency_ms=
#                                           row['max_latency_ms'],
#                                           comment=row['comment'])
#    finally:
#        connection.close()
#
#
#    # We try to register again and check that registration is idempotent
#
#    proxysql_instance.register_backend(mysql_backend_writer)
#    proxysql_instance.register_backend(mysql_backend_reader)
#
#
#    assert proxysql_instance.find_backends(hostgroup_writer)[0].hostname == backend_writer.hostname  # NOQA
#    assert proxysql_instance.find_backends(hostgroup_reader)[0].hostname == backend_reader.hostname  # NOQA



"""Module describes GaleraCluster class"""
from pymysql import OperationalError

from proxysql_tools import LOG
from proxysql_tools.galera.exceptions import GaleraClusterSyncedNodeNotFound, \
    GaleraClusterNodeNotFound
from proxysql_tools.galera.galera_node import GaleraNode, GaleraNodeState


class GaleraCluster(object):
    """
    GaleraCluster describes Galera cluster.

    :param cluster_hosts: .
    :type cluster_hosts: str
    :param user: MySQL user to connect to a cluster node.
    :type user: str
    :param password: MySQL password.
    :type password: str
    """
    def __init__(self, cluster_hosts, user='root', password=None):
        self._nodes = []
        for host in self._split_cluster_host(cluster_hosts):
            self._nodes.append(GaleraNode(host=host[0], port=host[1],
                                          user=user, password=password))

    @property
    def nodes(self):
        """
        Get list of Galera nodes

        :return: Return list of Galera nodes
        :rtype: list(GaleraNode)"""
        return self._nodes

    def find_node(self, host, port):
        """
        BY given host and port find a node in the cluster.

        :param host: IP or address of node.
        :param port: node port.
        :return: GaleraNode instance.
        :rtype: GaleraNode
        :raise: GaleraClusterNodeNotFound
        """
        for node in self._nodes:
            if node.host == host and node.port == port:
                return node

        raise GaleraClusterNodeNotFound('Cannot find node %s:%d', host, port)

    def find_synced_nodes(self):
        """Find a node in the cluster in SYNCED state.
        :return: List of Galera node in SYNCED state.
        :rtype: list(GaleraNode)
        :raise: GaleraClusterSyncedNodeNotFound
        """
        LOG.debug('Looking for a SYNCED node')
        nodes = []
        for galera_node in self._nodes:
            try:
                state = galera_node.wsrep_local_state
                LOG.debug('%s state: %s', galera_node, state)
                if state == GaleraNodeState.SYNCED:
                    nodes.append(galera_node)
            except OperationalError as err:
                LOG.error(err)
                LOG.info('Skipping node %s', galera_node)
        if nodes:
            return nodes
        else:
            raise GaleraClusterSyncedNodeNotFound('Cluster has '
                                                  'no SYNCED nodes')

    @staticmethod
    def _split_cluster_host(cluster_host):
        """Split a string with list of hosts and make a list of tuples out of it.
        For example, string
        *192.168.90.2:3306,192.168.90.3:3306,192.168.90.4:3306*
        will be converted into list:

.. code-block:: python

    [
        (192.168.90.2, 3306),
        (192.168.90.3, 3306),
        (192.168.90.4, 3306),
    ]


:param cluster_host: String with list of host:port pairs
:type cluster_host: str
:return: list of tuples (host, port)
:rtype: list(tuple)
        """
        result = []
        for item in cluster_host.split(','):
            host, port = item.split(':')
            result.append((host, int(port)))

        return result



import mock
import pytest
from pymysql import OperationalError
from pymysql.cursors import DictCursor

from proxysql_tools.proxysql.exceptions import ProxySQLBackendNotFound
from proxysql_tools.proxysql.proxysql import BackendStatus, ProxySQLMySQLBackend, \
    ProxySQLMySQLUser, ProxySQL


def test_backendstatus():
    bs = BackendStatus()
    assert bs.online == 'ONLINE'
    assert bs.shunned == 'SHUNNED'
    assert bs.offline_hard == 'OFFLINE_HARD'
    assert bs.offline_soft == 'OFFLINE_SOFT'


def test_proxysql_mysql_backend():
    be = ProxySQLMySQLBackend('foo',
                              hostgroup_id='0',
                              port='3307',
                              status=BackendStatus.offline_hard,
                              weight='1',
                              compression='1',
                              max_connections='10',
                              max_replication_lag='20',
                              use_ssl=100,
                              max_latency_ms='30',
                              comment='bar')

    assert be.hostgroup_id == 0
    assert be.port == 3307
    assert be.status == 'OFFLINE_HARD'
    assert be.weight == 1
    assert be.compression == 1
    assert be.max_connections == 10
    assert be.max_replication_lag == 20
    assert be.use_ssl is True
    assert be.max_latency_ms == 30
    assert be.comment == 'bar'


def test_proxysql_mysql_user():
    mu = ProxySQLMySQLUser(user='foo',
                           password='qwerty',
                           active=True,
                           use_ssl=True,
                           default_hostgroup='10',
                           default_schema='bar',
                           schema_locked=True,
                           transaction_persistent=True,
                           fast_forward=True,
                           backend=True,
                           frontend=False,
                           max_connections='10')
    assert mu.user == 'foo'
    assert mu.password == 'qwerty'
    assert mu.active is True
    assert mu.use_ssl is True
    assert mu.default_hostgroup == 10
    assert mu.default_schema == 'bar'
    assert mu.schema_locked is True
    assert mu.transaction_persistent is True
    assert mu.fast_forward is True
    assert mu.backend is True
    assert mu.frontend is False
    assert mu.max_connections == 10


def test_proxysql():
    ps = ProxySQL(host='foo',
                  port='3307',
                  user='bar',
                  password='qwerty')
    assert ps.host == 'foo'
    assert ps.port == 3307
    assert ps.user == 'bar'
    assert ps.password == 'qwerty'


@mock.patch.object(ProxySQL, '_connect')
def test_execute(mock_connect, proxysql):
    """
    :param proxysql: ProxySQL instance
    :type proxysql: ProxySQL
    """
    mock_connection = mock.Mock()
    mock_connect.return_value.__enter__.return_value = mock_connection
    mock_cursor = mock.Mock()
    mock_connection.cursor.return_value = mock_cursor

    proxysql.execute('foo')
    mock_connect.assert_called_once_with()
    mock_connection.cursor.assert_called_once_with()
    mock_cursor.execute.assert_called_once_with('foo')
    mock_cursor.fetchall.assert_called_once_with()


@mock.patch.object(ProxySQL, 'execute')
def test_ping_alive(mock_execute, proxysql):
    mock_execute.return_value = [{
        'result': '1'
    }]
    assert proxysql.ping() is True
    mock_execute.assert_called_once_with('SELECT 1 AS result')


@mock.patch.object(ProxySQL, 'execute')
def test_ping_dead(mock_execute, proxysql):
    mock_execute.side_effect = OperationalError
    assert proxysql.ping() is False
    mock_execute.assert_called_once_with('SELECT 1 AS result')


@mock.patch.object(ProxySQL, 'execute')
def test_reload_runtime(mock_execute, proxysql):
    """
    :param proxysql: ProxySQL instance
    :type proxysql: ProxySQL
    """
    proxysql.reload_runtime()
    calls = [mock.call('LOAD MYSQL SERVERS TO RUNTIME'),
             mock.call('LOAD MYSQL USERS TO RUNTIME'),
             mock.call('LOAD MYSQL VARIABLES TO RUNTIME')]
    mock_execute.assert_has_calls(calls=calls, any_order=False)


@pytest.mark.parametrize('comment, query',[
    (
        'Some comment',
        "REPLACE INTO mysql_servers(`hostgroup_id`, `hostname`, `port`, `status`, `weight`, `compression`, `max_connections`, `max_replication_lag`, `use_ssl`, `max_latency_ms`, `comment`) VALUES(0, 'foo', 3306, 'ONLINE', 1, 0, 10000, 0, 0, 0, 'Some comment')"
    ),
    (
        None,
        "REPLACE INTO mysql_servers(`hostgroup_id`, `hostname`, `port`, `status`, `weight`, `compression`, `max_connections`, `max_replication_lag`, `use_ssl`, `max_latency_ms`, `comment`) VALUES(0, 'foo', 3306, 'ONLINE', 1, 0, 10000, 0, 0, 0, NULL)"
    )
])
@mock.patch.object(ProxySQL, 'reload_runtime')
@mock.patch.object(ProxySQL, 'execute')
def test_register_backend(mock_execute, mock_runtime, comment, query, proxysql):
    """

    :param mock_execute:
    :param mock_runtime:
    :param comment:
    :param query:
    :param proxysql:
    :type proxysql: ProxySQL
    """
    backend = ProxySQLMySQLBackend('foo', comment=comment)
    proxysql.register_backend(backend)
    mock_execute.assert_called_once_with(query)
    mock_runtime.assert_called_once_with()


@mock.patch.object(ProxySQL, 'reload_runtime')
@mock.patch.object(ProxySQL, 'execute')
def test_deregister_backend(mock_execute, mock_runtime, proxysql):
    """

    :param mock_execute:
    :param mock_runtime:
    :param comment:
    :param query:
    :param proxysql:
    :type proxysql: ProxySQL
    """
    backend = ProxySQLMySQLBackend('foo', hostgroup_id=10, port=3307)
    proxysql.deregister_backend(backend)
    query = "DELETE FROM mysql_servers WHERE hostgroup_id=10 AND hostname='foo' AND port=3307"
    mock_execute.assert_called_once_with(query)
    mock_runtime.assert_called_once_with()


@pytest.mark.parametrize('kwargs_in, kwargs_out', [
    (
        {
            'socket': '/foo/bar'
        },
        {
            'connect_timeout': 20,
            'cursorclass': DictCursor,
            'passwd': None,
            'unix_socket': '/foo/bar',
            'user': 'root'

        }
    ),
    (
        {
            'port': 3307
        },
        {
            'connect_timeout': 20,
            'cursorclass': DictCursor,
            'passwd': None,
            'host': 'localhost',
            'port': 3307,
            'user': 'root'

        }
    ),
    (
        {

        },
        {
            'connect_timeout': 20,
            'cursorclass': DictCursor,
            'passwd': None,
            'host': 'localhost',
            'port': 3306,
            'user': 'root'

        }
    )
])
@mock.patch('proxysql_tools.proxysql.proxysql.pymysql')
def test_connect(mock_pymysql, kwargs_in, kwargs_out):
    ps = ProxySQL(**kwargs_in)
    with ps._connect() as conn:
        mock_pymysql.connect.assert_called_once_with(**kwargs_out)


@pytest.mark.parametrize('response, backend',[
    (
        [{
            u'status': 'ONLINE',
            u'comment': '',
            u'compression': '0',
            u'weight': '1',
            u'hostname': '192.168.90.2',
            u'hostgroup_id': '10',
            u'use_ssl': '0',
            u'max_replication_lag': '0',
            u'port': '3306',
            u'max_latency_ms': '0',
            u'max_connections': '10000'
        }]
        ,
        ProxySQLMySQLBackend('192.168.90.2', hostgroup_id=10, port=3306)
    )
])
@mock.patch.object(ProxySQL, 'execute')
def test_find_backends(mock_execute, proxysql, response, backend):
    mock_execute.return_value = response
    assert proxysql.find_backends(10)[0] == backend


@mock.patch.object(ProxySQL, 'execute')
def test_find_backends_raises(mock_execute, proxysql):
    mock_execute.return_value = ()
    with pytest.raises(ProxySQLBackendNotFound):
        proxysql.find_backends(10)



"""Entry points for proxysql-tools"""
from __future__ import print_function
import os
from ConfigParser import ConfigParser, NoOptionError, NoSectionError

import click
from pymysql import OperationalError

from proxysql_tools import setup_logging, LOG, __version__
from proxysql_tools.aws.aws import aws_notify_master
from proxysql_tools.cli_entrypoint.galera import galera_register
from proxysql_tools.proxysql.proxysql import ProxySQL
from proxysql_tools.util.bug1258464 import bug1258464

PASS_CFG = click.make_pass_decorator(ConfigParser, ensure=True)


@click.group(invoke_without_command=True)
@click.option('--debug', help='Print debug messages', is_flag=True,
              default=False)
@click.option('--config', help='ProxySQL Tools configuration file.',
              default='/etc/twindb/proxysql-tools.cfg',
              show_default=True)
@click.option('--version', help='Show tool version and exit.', is_flag=True,
              default=False)
@PASS_CFG
@click.pass_context
def main(ctx, cfg, debug, config, version):
    """proxysql-tool entrypoint"""
    if ctx.invoked_subcommand is None:
        if version:
            print(__version__)
            exit(0)
        else:
            print(ctx.get_help())
            exit(-1)

    setup_logging(LOG, debug=debug)

    if os.path.exists(config):
        cfg.read(config)
    else:
        LOG.error("Config file %s doesn't exist", config)
        exit(1)


@main.command()
@PASS_CFG
def ping(cfg):
    """Checks the health of ProxySQL."""
    kwargs = {}
    option_mapping = {
        'host': 'host',
        'port': 'admin_port',
        'user': 'admin_username',
        'password': 'admin_password'
    }

    for key in option_mapping:
        try:
            kwargs[key] = cfg.get('proxysql', option_mapping[key])
        except NoOptionError:
            pass

    if ProxySQL(**kwargs).ping():
        LOG.debug('ProxySQL is alive')
        exit(0)
    else:
        LOG.debug('ProxySQL is dead')
        exit(1)


@main.group()
def aws():
    """Commands to interact with ProxySQL on AWS."""


@aws.command()
@PASS_CFG
def notify_master(cfg):
    """The notify_master script for keepalived."""
    LOG.debug('Switching to master role and executing keepalived '
              'notify_master script.')
    aws_notify_master(cfg)


@main.group()
def galera():
    """Commands for ProxySQL and Galera integration."""


@galera.command()
@PASS_CFG
def register(cfg):
    """Registers Galera cluster nodes with ProxySQL."""

    try:
        galera_register(cfg)
    except NotImplementedError as err:
        LOG.error(err)
        exit(1)
    except (NoOptionError, NoSectionError) as err:
        LOG.error('Failed to parse config: %s', err)
        exit(1)

    except OperationalError as err:
        LOG.error('Failed to talk to database: %s', err)


@galera.command()
@PASS_CFG
@click.option('--default-file', help='Path to my.cnf with custom settings')
def bug1258464killer(default_file):
    """
    bug1258464killer checks status of a local Galera node
    and if a) There are stuck COMMIT queries and b) There is an ALTER TABLE
    it will kill the node. This command workarounds a known bug
    https://bugs.launchpad.net/percona-xtradb-cluster/+bug/1258464
    """
    if os.path.isfile(default_file):
        bug1258464(default_file)
    else:
        LOG.error("Config file %s doesn't exist", default_file)



from click.testing import CliRunner

import proxysql_tools
from proxysql_tools.cli import main
from proxysql_tools.galera.galera_node import GaleraNode
from proxysql_tools.proxysql.proxysql import ProxySQLMySQLBackend, BackendStatus
from tests.integration.library import proxysql_tools_config, \
    wait_for_cluster_nodes_to_become_healthy, proxysql_tools_config_2, \
    shutdown_container
import pymysql
from pymysql.cursors import DictCursor


def test__main_command_version_can_be_fetched():
    runner = CliRunner()
    result = runner.invoke(main, ['--version'])
    assert result.output == proxysql_tools.__version__ + '\n'
    assert result.exit_code == 0


def test__ping_command_can_be_executed(proxysql_instance, tmpdir):
    config = proxysql_tools_config(proxysql_instance, '127.0.0.1', '3306',
                                   'user', 'pass', 10, 11, 'monitor',
                                   'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)
        proxysql_tools.LOG.debug('proxysql-tools config: \n%s', config)
    runner = CliRunner()
    result = runner.invoke(main, ['--config', config_file, 'ping'])
    assert result.exit_code == 0


def test__galera_register_command_set_nodes_online(percona_xtradb_cluster_three_node,
                                                   proxysql_instance,
                                                   tmpdir):
    wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_three_node)
    hostgroup_writer = 10
    hostgroup_reader = 11

    rw_map = {
        0: hostgroup_writer,
        1: hostgroup_reader,
        2: hostgroup_reader
    }
    for i in xrange(3):
        backend = ProxySQLMySQLBackend(
            hostname=percona_xtradb_cluster_three_node[i]['ip'],
            port=percona_xtradb_cluster_three_node[0]['mysql_port'],
            hostgroup_id=rw_map[i]
        )
        proxysql_instance.register_backend(backend)

    blacklist = '{}:3306'.format(percona_xtradb_cluster_three_node[2]['ip'])
    nodes = [
        percona_xtradb_cluster_three_node[0]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[1]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[2]['ip'] + ':3306'
    ]
    config = proxysql_tools_config_2(proxysql_instance,
                                     nodes,
                                     'root', 'r00t', hostgroup_writer,
                                     hostgroup_reader,
                                     blacklist, 'monitor',
                                     'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)
        proxysql_tools.LOG.debug('proxysql-tools config: \n%s', config)
    runner = CliRunner()
    result = runner.invoke(main,
                           ['--config', config_file, 'galera', 'register']
                           )
    assert result.exit_code == 0

    connection = pymysql.connect(
        host=proxysql_instance.host,
        port=proxysql_instance.port,
        user=proxysql_instance.user,
        passwd=proxysql_instance.password,
        connect_timeout=20,
        cursorclass=DictCursor)
    try:
        with connection.cursor() as cursor:
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s', hostgroup_writer)
            for row in cursor.fetchall():
                backend = ProxySQLMySQLBackend(row['hostname'],
                                               hostgroup_id=row['hostgroup_id'],
                                               port=row['port'],
                                               status=row['status'],
                                               weight=row['weight'],
                                               compression=row['compression'],
                                               max_connections=
                                               row['max_connections'],
                                               max_replication_lag=
                                               row['max_replication_lag'],
                                               use_ssl=row['use_ssl'],
                                               max_latency_ms=
                                               row['max_latency_ms'],
                                               comment=row['comment'])
                assert backend.status == BackendStatus.online

            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s', hostgroup_reader)
            for row in cursor.fetchall():
                backend = ProxySQLMySQLBackend(row['hostname'],
                                               hostgroup_id=row['hostgroup_id'],
                                               port=row['port'],
                                               status=row['status'],
                                               weight=row['weight'],
                                               compression=row['compression'],
                                               max_connections=
                                               row['max_connections'],
                                               max_replication_lag=
                                               row['max_replication_lag'],
                                               use_ssl=row['use_ssl'],
                                               max_latency_ms=
                                               row['max_latency_ms'],
                                               comment=row['comment'])
                assert backend.status == BackendStatus.online

    finally:
        connection.close()


def test__galera_register_shutdowned_reader_is_degistered(percona_xtradb_cluster_three_node,
                                                          proxysql_instance,
                                                          tmpdir):
    wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_three_node)
    hostgroup_writer = 10
    hostgroup_reader = 11

    backend = ProxySQLMySQLBackend(
        hostname=percona_xtradb_cluster_three_node[0]['ip'],
        port=percona_xtradb_cluster_three_node[0]['mysql_port'],
        hostgroup_id=hostgroup_writer
    )
    proxysql_instance.register_backend(backend)

    backend_unreg = ProxySQLMySQLBackend(
        hostname=percona_xtradb_cluster_three_node[1]['ip'],
        port=percona_xtradb_cluster_three_node[1]['mysql_port'],
        hostgroup_id=hostgroup_reader
    )
    proxysql_instance.register_backend(backend_unreg)

    backend = ProxySQLMySQLBackend(
        hostname=percona_xtradb_cluster_three_node[2]['ip'],
        port=percona_xtradb_cluster_three_node[2]['mysql_port'],
        hostgroup_id=hostgroup_reader
    )
    proxysql_instance.register_backend(backend)

    blacklist = '{}:3306'.format(percona_xtradb_cluster_three_node[2]['ip'])
    nodes = [
        percona_xtradb_cluster_three_node[0]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[1]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[2]['ip'] + ':3306'
    ]

    shutdown_container(percona_xtradb_cluster_three_node[1]['id'])
    config = proxysql_tools_config_2(proxysql_instance,
                                     nodes,
                                     'root', 'r00t', hostgroup_writer,
                                     hostgroup_reader,
                                     blacklist, 'monitor',
                                     'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)
        proxysql_tools.LOG.debug('proxysql-tools config: \n%s', config)
    runner = CliRunner()
    result = runner.invoke(main,
                           ['--config', config_file, 'galera', 'register']
                           )
    assert result.exit_code == 0

    connection = pymysql.connect(
        host=proxysql_instance.host,
        port=proxysql_instance.port,
        user=proxysql_instance.user,
        passwd=proxysql_instance.password,
        connect_timeout=20,
        cursorclass=DictCursor)
    try:
        with connection.cursor() as cursor:
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s '
                           ' AND `hostname` = %s '
                           ' AND `port` = %s',
                           (
                               backend_unreg.hostgroup_id,
                               backend_unreg.hostname,
                               backend_unreg.port
                           ))
            assert cursor.fetchall() == ()

    finally:
        connection.close()


def test__galera_register_shutdowned_writer_is_deregistered(percona_xtradb_cluster_three_node,
                                                            proxysql_instance,
                                                            tmpdir):
    wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_three_node)
    hostgroup_writer = 10
    hostgroup_reader = 11

    backend_unreg = ProxySQLMySQLBackend(
        hostname=percona_xtradb_cluster_three_node[0]['ip'],
        port=percona_xtradb_cluster_three_node[0]['mysql_port'],
        hostgroup_id=hostgroup_writer
    )
    proxysql_instance.register_backend(backend_unreg)

    backend = ProxySQLMySQLBackend(
        hostname=percona_xtradb_cluster_three_node[1]['ip'],
        port=percona_xtradb_cluster_three_node[1]['mysql_port'],
        hostgroup_id=hostgroup_reader
    )
    proxysql_instance.register_backend(backend)

    backend = ProxySQLMySQLBackend(
        hostname=percona_xtradb_cluster_three_node[2]['ip'],
        port=percona_xtradb_cluster_three_node[2]['mysql_port'],
        hostgroup_id=hostgroup_reader
    )
    proxysql_instance.register_backend(backend)

    blacklist = '{}:3306'.format(percona_xtradb_cluster_three_node[2]['ip'])
    nodes = [
        percona_xtradb_cluster_three_node[0]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[1]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[2]['ip'] + ':3306'
    ]

    shutdown_container(percona_xtradb_cluster_three_node[0]['id'])
    config = proxysql_tools_config_2(proxysql_instance,
                                     nodes,
                                     'root', 'r00t', hostgroup_writer,
                                     hostgroup_reader,
                                     blacklist, 'monitor',
                                     'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)
        proxysql_tools.LOG.debug('proxysql-tools config: \n%s', config)
    runner = CliRunner()
    result = runner.invoke(main,
                           ['--config', config_file, 'galera', 'register']
                           )
    assert result.exit_code == 0

    connection = pymysql.connect(
        host=proxysql_instance.host,
        port=proxysql_instance.port,
        user=proxysql_instance.user,
        passwd=proxysql_instance.password,
        connect_timeout=20,
        cursorclass=DictCursor
    )
    try:
        with connection.cursor() as cursor:
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s '
                           ' AND `hostname` = %s '
                           ' AND `port` = %s',
                           (
                               backend_unreg.hostgroup_id,
                               backend_unreg.hostname,
                               backend_unreg.port
                           )
                           )
            assert cursor.fetchall() == ()

            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s ',
                           (
                               backend_unreg.hostgroup_id,
                           )
                           )
            assert cursor.fetchall() != ()

    finally:
        connection.close()


def test__galera_register_sync_desync_state(percona_xtradb_cluster_three_node,
                                                   proxysql_instance,
                                                   tmpdir):
    wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_three_node)
    hostgroup_writer = 10
    hostgroup_reader = 11

    rw_map = {
        0: hostgroup_writer,
        1: hostgroup_reader,
        2: hostgroup_reader
    }
    for i in xrange(3):
        backend = ProxySQLMySQLBackend(
            hostname=percona_xtradb_cluster_three_node[i]['ip'],
            port=percona_xtradb_cluster_three_node[0]['mysql_port'],
            hostgroup_id=rw_map[i]
        )
        proxysql_instance.register_backend(backend)


    desync_node = GaleraNode(
        host=percona_xtradb_cluster_three_node[1]['ip'],
        port=percona_xtradb_cluster_three_node[1]['mysql_port'],
        user='root',password='r00t'
    )
    desync_node.execute('set global wsrep_desync=ON;')

    blacklist = '{}:3306'.format(percona_xtradb_cluster_three_node[2]['ip'])
    nodes = [
        percona_xtradb_cluster_three_node[0]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[1]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[2]['ip'] + ':3306'
    ]
    config = proxysql_tools_config_2(proxysql_instance,
                                     nodes,
                                     'root', 'r00t', hostgroup_writer,
                                     hostgroup_reader,
                                     blacklist, 'monitor',
                                     'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)
        proxysql_tools.LOG.debug('proxysql-tools config: \n%s', config)
    runner = CliRunner()
    result = runner.invoke(main,
                           ['--config', config_file, 'galera', 'register'])
    assert result.exit_code == 0

    connection = pymysql.connect(
        host=proxysql_instance.host,
        port=proxysql_instance.port,
        user=proxysql_instance.user,
        passwd=proxysql_instance.password,
        connect_timeout=20,
        cursorclass=DictCursor)
    try:
        with connection.cursor() as cursor:
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s'
                           ' AND hostname = %s',
                           (
                               hostgroup_reader,
                               percona_xtradb_cluster_three_node[1]['ip']
                           )
                           )
            row = cursor.fetchall()[0]
            assert row['status'] == BackendStatus.offline_soft
    finally:
        connection.close()

    desync_node.execute('set global wsrep_desync=OFF;')
    runner = CliRunner()
    result = runner.invoke(main,
                           ['--config', config_file, 'galera', 'register'])
    assert result.exit_code == 0
    connection = pymysql.connect(
        host=proxysql_instance.host,
        port=proxysql_instance.port,
        user=proxysql_instance.user,
        passwd=proxysql_instance.password,
        connect_timeout=20,
        cursorclass=DictCursor)
    try:
        with connection.cursor() as cursor:
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s'
                           ' AND hostname = %s',
                           (
                               hostgroup_reader,
                               percona_xtradb_cluster_three_node[1]['ip']
                           )
                           )
            row = cursor.fetchall()[0]
            assert row['status'] == BackendStatus.online
    finally:
        connection.close()


def test__galera_register_writer_node_is_reader_when_readers_list_is_empty(percona_xtradb_cluster_three_node,
                                                   proxysql_instance,
                                                   tmpdir):

    wait_for_cluster_nodes_to_become_healthy(percona_xtradb_cluster_three_node)
    hostgroup_writer = 10
    hostgroup_reader = 11

    rw_map = {
        0: hostgroup_writer,
        1: hostgroup_reader,
        2: hostgroup_reader
    }
    for i in xrange(3):
        backend = ProxySQLMySQLBackend(
            hostname=percona_xtradb_cluster_three_node[i]['ip'],
            port=percona_xtradb_cluster_three_node[0]['mysql_port'],
            hostgroup_id=rw_map[i]
        )
        proxysql_instance.register_backend(backend)


    desync_node = GaleraNode(
        host=percona_xtradb_cluster_three_node[1]['ip'],
        port=percona_xtradb_cluster_three_node[1]['mysql_port'],
        user='root',password='r00t'
    )
    desync_node.execute('set global wsrep_desync=ON;')

    desync_node = GaleraNode(
        host=percona_xtradb_cluster_three_node[2]['ip'],
        port=percona_xtradb_cluster_three_node[2]['mysql_port'],
        user='root',password='r00t'
    )
    desync_node.execute('set global wsrep_desync=ON;')

    blacklist = '{}:3306'.format(percona_xtradb_cluster_three_node[2]['ip'])
    nodes = [
        percona_xtradb_cluster_three_node[0]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[1]['ip'] + ':3306',
        percona_xtradb_cluster_three_node[2]['ip'] + ':3306'
    ]
    config = proxysql_tools_config_2(proxysql_instance,
                                     nodes,
                                     'root', 'r00t', hostgroup_writer,
                                     hostgroup_reader,
                                     blacklist, 'monitor',
                                     'monitor')
    config_file = str(tmpdir.join('proxysql-tool.cfg'))
    with open(config_file, 'w') as fh:
        config.write(fh)
        proxysql_tools.LOG.debug('proxysql-tools config: \n%s', config)
    runner = CliRunner()
    result = runner.invoke(main,
                           ['--config', config_file, 'galera', 'register'])
    assert result.exit_code == 0

    connection = pymysql.connect(
        host=proxysql_instance.host,
        port=proxysql_instance.port,
        user=proxysql_instance.user,
        passwd=proxysql_instance.password,
        connect_timeout=20,
        cursorclass=DictCursor)
    try:
        with connection.cursor() as cursor:
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s'
                           ' AND hostname = %s',
                           (
                               hostgroup_reader,
                               percona_xtradb_cluster_three_node[1]['ip']
                           )
                           )
            row = cursor.fetchall()[0]
            assert row['status'] == BackendStatus.offline_soft
            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s'
                           ' AND hostname = %s',
                           (
                               hostgroup_reader,
                               percona_xtradb_cluster_three_node[2]['ip']
                           )
                           )
            row = cursor.fetchall()[0]
            assert row['status'] == BackendStatus.offline_soft

            cursor.execute('SELECT `hostgroup_id`, `hostname`, '
                           '`port`, `status`, `weight`, `compression`, '
                           '`max_connections`, `max_replication_lag`, '
                           '`use_ssl`, `max_latency_ms`, `comment`'
                           ' FROM `mysql_servers`'
                           ' WHERE hostgroup_id = %s'
                           ' AND hostname = %s',
                           (
                               hostgroup_reader,
                               percona_xtradb_cluster_three_node[0]['ip']
                           )
                           )
            assert cursor.fetchall() != ()

    finally:
        connection.close()

